// @ts-check

/**
 * ==============================================================================
 * Imports
 * ==============================================================================
 */
const path = require("path");
const fs = require("fs");
const https = require("https");
const http = require("http");
const grabHostNames = require("../package-shared/utils/grab-host-names");
const apiGetUser = require("../package-shared/functions/api/users/api-get-user");

/** ****************************************************************************** */
/** ****************************************************************************** */
/** ****************************************************************************** */
/** ****************************************************************************** */
/** ****************************************************************************** */
/** ****************************************************************************** */

/**
 * ==============================================================================
 * Main Function
 * ==============================================================================
 * @async
 *
 * @param {object} params - Single Param object containing params
 * @param {String} params.key - API Key
 * @param {String} params.database - Target Database
 * @param {number} params.userId - user to get
 * @param {string[]} [params.fields] - fields to select
 * @param {boolean} [params.apiUserId] - API User ID
 * @param {boolean} [params.useLocal] - Whether to use a remote database instead of API
 *
 * @returns { Promise<import("../package-shared/types").GetUserFunctionReturn>}
 */
async function getUser({ key, userId, database, fields, apiUserId, useLocal }) {
    /**
     * Initialize
     */
    const defaultFields = [
        "id",
        "first_name",
        "last_name",
        "email",
        "username",
        "image",
        "image_thumbnail",
        "verification_status",
        "date_created",
        "date_created_code",
        "date_created_timestamp",
        "date_updated",
        "date_updated_code",
        "date_updated_timestamp",
    ];

    const updatedFields =
        fields && fields[0] ? [...defaultFields, ...fields] : defaultFields;

    const reqPayload = JSON.stringify({
        userId,
        database,
        fields: [...new Set(updatedFields)],
    });

    const grabedHostNames = grabHostNames();
    const { host, port, scheme } = grabedHostNames;

    /**
     * Check for local DB settings
     *
     * @description Look for local db settings in `.env` file and by pass the http request if available
     */
    const { DSQL_DB_HOST, DSQL_DB_USERNAME, DSQL_DB_PASSWORD, DSQL_DB_NAME } =
        process.env;

    if (
        DSQL_DB_HOST?.match(/./) &&
        DSQL_DB_USERNAME?.match(/./) &&
        DSQL_DB_PASSWORD?.match(/./) &&
        DSQL_DB_NAME?.match(/./) &&
        useLocal
    ) {
        /** @type {import("../package-shared/types").DSQL_DatabaseSchemaType | undefined} */
        let dbSchema;

        try {
            const localDbSchemaPath = path.resolve(
                process.cwd(),
                "dsql.schema.json"
            );
            dbSchema = JSON.parse(fs.readFileSync(localDbSchemaPath, "utf8"));
        } catch (error) {}

        return await apiGetUser({
            userId,
            fields: [...new Set(updatedFields)],
            dbFullName: DSQL_DB_NAME,
            useLocal,
        });
    }

    /**
     * Make https request
     *
     * @description make a request to datasquirel.com
     */
    const httpResponse = await new Promise((resolve, reject) => {
        const httpsRequest = scheme.request(
            {
                method: "POST",
                headers: {
                    "Content-Type": "application/json",
                    "Content-Length": Buffer.from(reqPayload).length,
                    Authorization:
                        key ||
                        process.env.DSQL_FULL_ACCESS_API_KEY ||
                        process.env.DSQL_API_KEY,
                },
                port,
                hostname: host,
                path: `/api/user/${
                    apiUserId || grabedHostNames.user_id
                }/get-user`,
            },

            /**
             * Callback Function
             *
             * @description https request callback
             */
            (response) => {
                var str = "";

                response.on("data", function (chunk) {
                    str += chunk;
                });

                response.on("end", function () {
                    resolve(JSON.parse(str));
                });

                response.on("error", (err) => {
                    reject(err);
                });
            }
        );

        httpsRequest.write(reqPayload);
        httpsRequest.end();
    });

    /** ********************************************** */
    /** ********************************************** */
    /** ********************************************** */

    return httpResponse;
}

/** ********************************************** */
/** ********************************************** */
/** ********************************************** */

module.exports = getUser;
